#include "AudioSystem.h"
#include <cmath>

const double Pi = 3.1415926535897932384626433832795;

SimpleSound::SimpleSound(float frequency, float volume, float panning)
{
	samplecount = 0;
	freq = frequency;
	freqshift = 0;
	vol = volume;
	volshift = -0.4f;
	pan = panning; // -1 == left, +1 = right;
	panshift = 0;
	wavepos = 0; // expressed in terms of cycles, not as an angle.

	active = true;
}
SimpleSound::~SimpleSound()
{
}
void SimpleSound::AddSound(int numsamples, int* samplebuffer)
{
	int i;

	float amp = 16384; // Amplitude adjustment (50% of possible volume)
	float sample;
	float mixleft, mixright;
	int sampleft, sampright;
	// Only update pan every call, but update frequency and volume continuously.
	mixleft = (float)cos(Pi*(pan+1)/4); // -1 == full left, +1 == full right
	mixright = (float)sin(Pi*(pan+1)/4);

	for(i=0;i<numsamples;i++)
	{
		sample = (float)sin(wavepos*2*Pi)*amp*vol;

		wavepos += freq/samplerate;
		freq += freqshift/samplerate;

		vol += volshift/samplerate;
		if(vol < 0) vol = 0;
		if(vol > 2) vol = 2;

		sampleft = (int)(sample*mixleft);
		sampright = (int)(sample*mixright);
		samplebuffer[i*2+0] += sampleft;
		samplebuffer[i*2+1] += sampright;
	}
	wavepos = wavepos - floor(wavepos); // Make wavepos smaller (don't lose as much precision)

	// Update & clamp pan value
	pan += numsamples*panshift / samplerate;
	if(pan < -1) pan = -1;
	if(pan > 1) pan = 1;

	// Drop sound if volume is 0;
	if(vol == 0) active=false;

	samplecount += numsamples;

}

/*
int samplecount;
// Shift values are shift per second.
float freq;
float freqshift;
float vol, volshift;
float pan, panshift;

float wavepos;
*/












// *2+0 = Left / *2+1 = Right






AudioSystem::AudioSystem(void)
{
	playing = false;
	error = false;

	int i;
	for(i=0;i<max_sounds;i++)
	{
		sounds[i]=NULL;
	}

	// Start rendering thread
	DWORD threadID = 0;
	hThread = CreateThread(NULL,0,AudioSystem::AudioThreadProc,this,CREATE_SUSPENDED,&threadID);
	if(hThread == INVALID_HANDLE_VALUE)
	{
		error = true;
	}
	// Initialize wave mapper device (native win32!)
	WAVEFORMATEX wfex;
	wfex.cbSize = sizeof(wfex);
	wfex.nAvgBytesPerSec=44100*2*2;
	wfex.nBlockAlign=4;
	wfex.nChannels=2;
	wfex.nSamplesPerSec=44100;
	wfex.wBitsPerSample=16;
	wfex.wFormatTag = WAVE_FORMAT_PCM;

	MMRESULT result = waveOutOpen(&hWO, WAVE_MAPPER, &wfex, threadID, (DWORD_PTR)this, CALLBACK_THREAD);
	if(result!=MMSYSERR_NOERROR)
	{
		error = true;
	}
	for(i=0;i<num_buffers;i++)
	{
		wh[i].dwBufferLength=buffer_samples*4;
		wh[i].dwBytesRecorded=0;
		wh[i].dwFlags=0;
		wh[i].dwLoops=0;
		wh[i].dwUser=i;
		wh[i].lpData=(LPSTR)buffers[i];
		wh[i].lpNext=0;
		wh[i].reserved=0;
		result = waveOutPrepareHeader(hWO,&wh[i],sizeof(wh[i]));
		if(result!=MMSYSERR_NOERROR)
		{
			error = true;
		}
	}

	// Setup exit signalling method
	hExitEvent = CreateEvent(NULL,TRUE,FALSE,NULL);
	hExitWaitEvent = CreateEvent(NULL,TRUE,FALSE,NULL);

	// Give thread a high priority & start it.
	SetThreadPriority(hThread,THREAD_PRIORITY_HIGHEST);
	ResumeThread(hThread);

}

AudioSystem::~AudioSystem(void)
{
	Stop(); // End playback

	SetEvent(hExitEvent); // Tell the thread to exit
	if(WAIT_OBJECT_0!=WaitForSingleObject(hExitWaitEvent,500)) // Give thread 500ms to exit
	{ // Otherwise kill it
		TerminateThread(hThread,0);
	}
	CloseHandle(hThread);
	CloseHandle(hExitEvent);
	CloseHandle(hExitWaitEvent);

	// Teardown wave mapper device
	int i;
	for(i=0;i<num_buffers;i++)
	{
		MMRESULT result = waveOutUnprepareHeader(hWO,&wh[i],sizeof(wh[i]));
		if(result!=MMSYSERR_NOERROR)
		{
			error = true;
		}
	}
	waveOutClose(hWO);
}

void AudioSystem::Start(void)
{
	int i;
	Stop();
	for(i=0;i<num_buffers;i++)
	{
		RenderBuffer(i);
	}
	playing = true;
	for(i=0;i<num_buffers;i++)
	{
		PlayBuffer(i);
	}
}
void AudioSystem::Stop(void)
{
	playing = false;
	waveOutReset(hWO);
}

void AudioSystem::PlaySnd(SoundEmitter * snd)
{
	int i;
	for(i=0;i<max_sounds;i++)
	{
		if(sounds[i] == NULL)
		{
			sounds[i] = snd;
			return;
		}
	}
}



void AudioSystem::PlayBuffer(int buffernum)
{
	MMRESULT result = waveOutWrite(hWO, &wh[buffernum], sizeof(wh[buffernum]));
	if(result != MMSYSERR_NOERROR)
	{
		error = true;
	}
}

void AudioSystem::RenderBuffer(int buffernum)
{
	int i;
	for(i=0;i<buffer_samples*2;i++)
	{
		mixbuffer[i] = 0; 
	}
	for(i=0;i<max_sounds;i++)
	{
		if(sounds[i] == NULL) continue;
		if(sounds[i]->active == false)
		{
			delete sounds[i]; // may cause problems... :)
			sounds[i] = NULL; 
			continue;
		}
		// Add this sound to the buffer.
		sounds[i]->AddSound(buffer_samples,mixbuffer);
	}

	for(i=0;i<buffer_samples*2;i++)
	{ // clip and output
		if(mixbuffer[i] > 32767) mixbuffer[i] = 32767;
		if(mixbuffer[i] < -32768) mixbuffer[i] = -32768;
		buffers[buffernum][i] = (short)mixbuffer[i];
	}
	// *2+0 = Left / *2+1 = Right
}
bool AudioSystem::IsPlaying()
{
	return playing;
}

DWORD WINAPI AudioSystem::AudioThreadProc(VOID* param)
{
	AudioSystem* a = (AudioSystem*) param;
	a->ThreadFunc();
	SetEvent(a->hExitWaitEvent);
	return 0;
}


void AudioSystem::ThreadFunc()
{
	while(true)
	{
		DWORD wait = MsgWaitForMultipleObjects(1,&hExitEvent,FALSE,INFINITE,QS_ALLPOSTMESSAGE);
		if(wait==0) 
		{ // Exit event signalled
			return;
		}
		if(wait!=1)
		{ // Hmm, returned an error?
			Sleep(2);
		}
		// Check for messages!
		MSG msg;
		while(PeekMessage(&msg,0,0,0,PM_REMOVE))
		{
			switch(msg.message)
			{
			case MM_WOM_DONE:
				if(IsPlaying())
				{
					WAVEHDR * wh = (WAVEHDR*)msg.lParam;
					int index = (int)wh->dwUser;
					RenderBuffer(index);
					if(!IsPlaying()) break; // Second chance; Try to decrease the chance of a race (without actually properly synchronizing :P)
					PlayBuffer(index);
				}
				break;
			}
		}
	}
}
