#pragma once

#include "Bitmap.h"
#include <cmath>
#include <vector>

struct Vertex
{
	float x,y,z;
};
struct VertexColor
{
	VertexColor() {} // Default: uninitialized
	VertexColor(float x_, float y_, float z_, float r_, float g_, float b_, float a_)
	{
		x=x_; y=y_; z=z_; r=r_; g=g_; b=b_; a=a_;
	}
	float x,y,z;
	float r,g,b,a;
};
struct VertexTexture
{
	float x,y,z;
	float u,v;
};
struct VertexColorTexture
{
	float x,y,z;
	float r,g,b,a;
	float u,v;
};
struct VertexNormal
{
	float x,y,z;
	float nx,ny,nz;
};
struct VertexNormalTexture
{
	float x,y,z;
	float u,v;
	float nx,ny,nz;
};
struct PixelColor
{
	float r,g,b,a;
};

inline unsigned long MakePx(PixelColor & px)
{ // Performs no bounds checking, for speed reasons
	return 
		(((int)(px.a*255))<<24) | 
		(((int)(px.r*255))<<16) | 
		(((int)(px.g*255))<<8) | 
		(((int)(px.b*255)));
}
inline void ReadPx(unsigned long px, PixelColor & dest)
{
	dest.a = ((px>>24)&255)/255.0f;
	dest.r = ((px>>16)&255)/255.0f;
	dest.g = ((px>>8 )&255)/255.0f;
	dest.b = ((px    )&255)/255.0f;
}

template<bool WrapU, bool WrapV>
inline void SamplePixelNearest(PixelColor & dest, Bitmap* texture, float u, float v)
{
	int iu,iv;
	if(WrapU) u = u-floor(u);
	if(WrapV) v = v-floor(u);
	iu = (int)u*texture->getWidth();
	iv = (int)u*texture->getHeight();
	ReadPx(texture->point(iu,iv).c,dest);
}

template<bool WrapU, bool WrapV>
inline void SamplePixelBilinear(PixelColor & dest, Bitmap* texture, float u, float v)
{
	SamplePixelNearest<WrapU,WrapV>(dest,texture,u,v); // bilinear not implemented yet.
}


template<typename VtxFormat>
void PixelShaderColor(VtxFormat & input, PixelColor & output, std::vector<Bitmap*>& textures)
{
	output.r = input.r;
	output.b = input.b;
	output.g = input.g;
	output.a = input.a;
}

template<typename VtxFormat>
void PixelShaderTexture(VtxFormat & input, PixelColor & output, std::vector<Bitmap*>& textures)
{
	SamplePixelBilinear<true,true>(output,textures[0],input.u, input.v);
}


inline void BlendModeOverwrite(unsigned long * pixel, PixelColor & pxcolor)
{
	*pixel = MakePx(pxcolor);
}
inline void BlendModeAlpha(unsigned long * pixel, PixelColor & pxcolor)
{
	if(pxcolor.a==0.0f) return; // very short circuit
	if(pxcolor.a==1.0f)
	{ // short circuit
		*pixel = MakePx(pxcolor);
	}
	else
	{ // Blend
		PixelColor px;
		float ia;
		ia = 1.0f - pxcolor.a;
		ReadPx(*pixel,px);

		px.a = pxcolor.a;
		px.r  = px.r * ia + pxcolor.r * pxcolor.a;
		px.g  = px.g * ia + pxcolor.g * pxcolor.a;
		px.b  = px.b * ia + pxcolor.b * pxcolor.a;

		*pixel = MakePx(pxcolor);	
	}
}
inline void BlendModeAdditive(unsigned long * pixel, PixelColor & pxcolor)
{
	PixelColor px;
	ReadPx(*pixel,px);
	px.a += pxcolor.a;
	px.r += pxcolor.r;
	px.g += pxcolor.g;
	px.b += pxcolor.b;
	// Bound check
	if(px.a > 1.0f) px.a = 1.0f;
	if(px.r > 1.0f) px.r = 1.0f;
	if(px.g > 1.0f) px.g = 1.0f;
	if(px.b > 1.0f) px.b = 1.0f;
	*pixel = MakePx(px);
}



template<bool PerspectiveCorrect = true, bool Clip = true>
class Rasterizer
{
protected:

	Bitmap * bmp;

public:
	Rasterizer() { bmp = 0; }
	~Rasterizer() {}

	void Bind(Bitmap * bind_bmp)
	{
		bmp = bind_bmp;
	}

	std::vector<Bitmap*> textures;


protected:

template<typename VtxFormat>
static inline void LerpCalcStep(VtxFormat & src, VtxFormat & dest, VtxFormat & step, float stepsize)
{
	const int icount = (sizeof(VtxFormat))/4;
	const int ioffset = 0;
	if(stepsize==0.0f) return; // Can't divide by 0.

	int i;
	for(i=ioffset;i<icount;i++)
	{
		((float*)&step)[i] = (((float*)&dest)[i]-((float*)&src)[i]) / stepsize;
	}
}

template<typename VtxFormat>
static inline void LerpCalcStepPerspective(VtxFormat & src, VtxFormat & dest, VtxFormat & step, float stepsize)
{ // This function may or may not create a perspective-correct step, depending on whether the PerspectiveCorrect flag is set.
	const int icount = (sizeof(VtxFormat))/4;
	const int ioffset = 0;
	const int pcoffset = 3;
	if(stepsize==0.0f) return; // Can't divide by 0.

	int i;
	if(PerspectiveCorrect)
	{
		for(i=ioffset;i<icount;i++)
		{
			if(i<pcoffset)
			{
				((float*)&step)[i] = (((float*)&dest)[i]-((float*)&src)[i]) / stepsize;
			}
			else
			{
				((float*)&step)[i] = (((float*)&dest)[i]/dest.z-((float*)&src)[i]/src.z) / stepsize;
			}
		}
	}
	else
	{
		for(i=ioffset;i<icount;i++)
		{
			((float*)&step)[i] = (((float*)&dest)[i]-((float*)&src)[i]) / stepsize;
		}
	}
}



template<typename VtxFormat>
static inline void LerpAddStep(VtxFormat & vtx, VtxFormat & step)
{
	const int icount = (sizeof(VtxFormat))/4;
	const int ioffset = 0;

	int i;
	for(i=ioffset;i<icount;i++)
	{
		((float*)&vtx)[i] += ((float*)&step)[i];
	}
}

template<typename VtxFormat>
static inline void LerpAddStep(VtxFormat & vtx, VtxFormat & step, float partialstep)
{
	const int icount = (sizeof(VtxFormat))/4;
	const int ioffset = 0;

	int i;
	for(i=ioffset;i<icount;i++)
	{
		((float*)&vtx)[i] += ((float*)&step)[i]*partialstep;
	}
}

template<typename VtxFormat>
static inline void LerpCreateStep(VtxFormat & source, VtxFormat & vtx, VtxFormat & step, float partialstep)
{
	const int icount = (sizeof(VtxFormat))/4;
	const int ioffset = 0;

	int i;
	for(i=ioffset;i<icount;i++)
	{
		((float*)&vtx)[i] = ((float*)&source)[i] + ((float*)&step)[i]*partialstep;
	}
}

template<typename VtxFormat>
static inline void LerpPerspectiveCorrect(VtxFormat & vtx)
{
	const int icount = (sizeof(VtxFormat))/4;
	const int ioffset = 3; // only perspective correct non-xyz entries.

	if(vtx.z == 0.0f) return; // Can't deal with this!
	int i;
	for(i=ioffset;i<icount;i++)
	{
		((float*)&vtx)[i] = ((float*)&vtx)[i]/vtx.z;
	}	
}

template<typename VtxFormat>
static inline void SortVertices(VtxFormat*&vtx1, VtxFormat*&vtx2, VtxFormat*&vtx3)
{
	// Sort vertices!
	VtxFormat * vtxtmp;
	if(vtx1->y <= vtx2->y)
	{
		if(vtx2->y <= vtx3->y)
		{
			// Correct order (1,2,3)
		}
		else
		{ // Vtx2 is the largest
			if(vtx1->y <= vtx3->y)
			{ // 1,3,2 -> 1,2,3 (2/3 swap)
				vtxtmp = vtx2; vtx2 = vtx3; vtx3 = vtxtmp;
			}
			else
			{ // 3,1,2 -> 1,2,3
				vtxtmp = vtx1; vtx1 = vtx3; vtx3 = vtx2; vtx2 = vtxtmp;
			}
		}
	}
	else
	{
		if(vtx1->y <= vtx3->y)
		{ // 2,1,3 (1/2 swap)
			vtxtmp = vtx1; vtx1 = vtx2; vtx2 = vtxtmp;
		}
		else
		{ 
			if(vtx2->y <= vtx3->y)
			{ // 2,3,1 -> 1,2,3
				vtxtmp = vtx2; vtx2 = vtx3; vtx3 = vtx1; vtx1 = vtxtmp;
			}
			else
			{ // 3,2,1 (1/3 swap)
				vtxtmp = vtx1; vtx1 = vtx3; vtx3 = vtxtmp;
			}
		}
	}
}


public:

template<typename VtxFormat, void PixelShaderFunc(VtxFormat &, PixelColor &, std::vector<Bitmap*>&), void BlendModeFunc(unsigned long *, PixelColor &)>
void RasterizeTriangleList(std::vector<VtxFormat> vertices)
{
	size_t i;
	for(i=0;i+2<vertices.size();i+=3)
	{
		Rasterize<VtxFormat,PixelShaderFunc,BlendModeFunc>(&vertices[i],&vertices[i+1],&vertices[i+2]);
	}
}


// The one that does it all!
template<typename VtxFormat, void PixelShaderFunc(VtxFormat &, PixelColor &, std::vector<Bitmap*>&), void BlendModeFunc(unsigned long *, PixelColor &)>
void Rasterize(VtxFormat * vtx1, VtxFormat * vtx2, VtxFormat * vtx3)
{
	const float px_x_offset = 0.5f;
	const float px_y_offset = 0.5f;

	SortVertices(vtx1,vtx2,vtx3);

	// Early elimination (Because these values will screw up the rest of the loop.)
	if(vtx3->y == vtx1->y) return;

	// Calculate left/right step values
	VtxFormat posLeft, posRight;
	VtxFormat stepLeft, stepRight;
	bool has_second_loop, update_left;
	float y,loop_end_y;

	y = vtx1->y;
	if(vtx1->y == vtx2->y)
	{ // flat top: special case.
		if(vtx1->x < vtx2->x)
		{
			posLeft = *vtx1;
			posRight = *vtx2;
		}
		else
		{
			posLeft = *vtx2;
			posRight = *vtx1;
		}
		LerpCalcStepPerspective(posLeft,*vtx3,stepLeft,vtx3->y-posLeft.y);
		LerpCalcStepPerspective(posRight,*vtx3,stepRight,vtx3->y-posLeft.y); // Can use Left for both, as we know left->y == right->y
		has_second_loop = false;
		loop_end_y = vtx3->y;
	}
	else
	{
		posLeft = *vtx1;
		posRight = *vtx1;
		// Determine if vtx1-vtx2 connection is the left or right side.
		if( (vtx2->x-vtx1->x)/(vtx2->y-vtx1->y) < (vtx3->x-vtx1->x)/(vtx3->y-vtx1->y) )
		{ // vtx2 on the Left side
			LerpCalcStepPerspective(posLeft,*vtx2,stepLeft,vtx2->y-posLeft.y);
			LerpCalcStepPerspective(posRight,*vtx3,stepRight,vtx3->y-posLeft.y); // Again, posleft/posright have the same y value
			update_left = true;
		}
		else
		{ // vtx2 on the Right side
			LerpCalcStepPerspective(posLeft,*vtx3,stepLeft,vtx3->y-posLeft.y);
			LerpCalcStepPerspective(posRight,*vtx2,stepRight,vtx2->y-posLeft.y); // posleft/posright have the same y value
			update_left = false;
		}
		has_second_loop = true;
		loop_end_y = vtx2->y; // second loop will go from vtx2 to vtx3 in the y axis.
	}




	{ // Rasterize loop
		VtxFormat curloc;
		VtxFormat stepHoriz;
		PixelColor pixel;
		float correct_y;
		float x, line_end_x;
		int int_x, int_y;

		unsigned long * bmpbits;
		int bmpwidth,bmpheight;

		bmpbits = bmp->getBits();
		bmpwidth = bmp->getWidth();
		bmpheight = bmp->getHeight();

		// Correct Y position to be pixel aligned to the next pixel
		correct_y = ceil(y-px_y_offset)- (y-px_y_offset); // Amount that needs to be added to Y to get a pixel-aligned value
		y+=correct_y;
		LerpAddStep(posLeft,stepLeft,correct_y);
		LerpAddStep(posRight,stepRight,correct_y);

		int_y = (int)(y-px_y_offset);

		if(Clip) // Clip against top edge
		{
			if(int_y < 0)
			{
				LerpAddStep(posLeft,stepLeft,(float)-int_y);
				LerpAddStep(posRight,stepRight,(float)-int_y);
				y-=int_y;
				int_y=0;
			}
		}

		while(true)
		{ // Outer loop (to allow the second loop)
			// Clip against bottom edge
			if(Clip)
			{
				if(loop_end_y>(bmpheight-1+px_y_offset)) loop_end_y = (bmpheight-1+px_y_offset+0.1f);
			}

			while(y<loop_end_y)
			{
				// Only render if span actually overlaps pixels
				if(posLeft.x < (floor(posRight.x - px_x_offset)+px_x_offset))
				{
					// Get horizontal span info
					LerpCalcStep(posLeft,posRight,stepHoriz,posRight.x-posLeft.x); // shouldn't calculate this with the perspective correction (already corrected)
					// Iterate over the pixels in the span
					line_end_x = posRight.x;
					x = posLeft.x;
					x = ceil(x-px_x_offset)+px_x_offset;
					int_x = (int)(x-px_x_offset);
					// Clip horizontal span
					if(Clip)
					{
						if(int_x < 0)
						{
							x-=int_x;
							int_x = 0;
						}
						if(line_end_x > (bmpwidth-1+px_x_offset)) line_end_x = (bmpwidth-1+px_x_offset+0.1f);
					}
					while(x < line_end_x)
					{
						LerpCreateStep(posLeft,curloc,stepHoriz,x-posLeft.x);
						if(PerspectiveCorrect) LerpPerspectiveCorrect(curloc);

						PixelShaderFunc(curloc,pixel,textures); // Run pixel shader

						BlendModeFunc(bmpbits+(int_x + int_y*bmpwidth),pixel); // Run blend mode function

						x += 1.0f;
						int_x++;
					}

				}
				// Prepare for the next step.
				int_y++; 
				y+=1.0f; 
				LerpAddStep(posLeft,stepLeft);
				LerpAddStep(posRight,stepRight);				
			}
			if(!has_second_loop) break;
			has_second_loop = false;
			correct_y = ceil(vtx2->y-px_y_offset)- (vtx2->y-px_y_offset); // Try to make the math as consistent between 1-2 and 2-3 edges as possible
			if(Clip)
			{
				int_y = (int)(vtx2->y + correct_y - px_y_offset);
			}
			if(update_left)
			{
				posLeft = *vtx2;
				LerpCalcStepPerspective(posLeft,*vtx3,stepLeft,vtx3->y-posLeft.y);
				LerpAddStep(posLeft,stepLeft,correct_y);
				if(Clip)
				{
					if(int_y<0)
					{
						LerpAddStep(posLeft,stepLeft,(float)-int_y);
						int_y=0;
					}
				}
			}
			else
			{
				posRight = *vtx2;
				LerpCalcStepPerspective(posRight,*vtx3,stepRight,vtx3->y-posRight.y);
				LerpAddStep(posRight,stepRight,correct_y);
				if(Clip)
				{
					if(int_y<0)
					{
						LerpAddStep(posRight,stepRight,(float)-int_y);
						int_y=0;
					}
				}
			}
			loop_end_y = vtx3->y;

		}
	}
}





};




